# see notes http://www.ics.uci.edu/~pattis/ICS-21/lectures/notes/Skip%20Lists.pdf

import Base.show
import Base.isempty
import Base.start
import Base.done
import Base.next

type Node
    key     ::  Number
    next    ::  Union(Node, Type{None})
    prev    ::  Union(Node, Type{None})
    below   ::  Union(Node, Type{None})
    above   ::  Union(Node, Type{None})

    Node(k::Number) = new(k, None, None, None, None)
end

type SkipList
    s   ::  Node # s is the start position of the Skip List
    h   ::  Int  # h is the height of the Skip List
    n   ::  Int  # n is the number of items in the Skip List

    SkipList() = begin
        head = Node(-Inf)
        tail = Node(Inf)

        head.next = tail
        tail.prev = head

        new(head, 0, 0)
    end
end

function search(skip::SkipList, k::Number)
    p = skip.s
    while p.below != None
        p = p.below
        while k >= p.next.key
            p = p.next
        end
    end
    p.key == k
end

function search(skip::SkipList, a::Number, b::Number)
    p = skip.s
    while p.below != None
        p = p.below
        while a >= p.next.key
            p = p.next
        end
    end
    A = Number[]
    while p.key <= b
        if p.key >= a
            append!(A, [p.key])
        end
        p = p.next
    end
    A
end

## General Collections ################
length(skip::SkipList) = skip.n
isempty(skip::SkipList) = skip.n == 0
#######################################

## Iterable Collections ###############
function start(skip::SkipList)
    s = skip.s
    while s.below != None
        s = s.below
    end
    s.next
end

function done(skip::SkipList, node)
    node.key == Inf
end

function next(skip::SkipList, node)
    node.key, node.next
end
#######################################


# insert the key k into the Skip List skip
function insert(skip::SkipList, k::Number)
    # returns a node p in the bottom of the Skip List such that
    # p.key is the largest key <= k
    function SkipSearch(skip::SkipList, k::Number)
        p = skip.s
        while p.below != None
            p = p.below
            while k >= p.next.key
                p = p.next
            end
        end
        p
    end

    # insertAfterAbove inserts a node with key k after the node p and above
    # above the node q; returns the new node r
    function insertAfterAbove(p::Union(Node, Type{None}),
                              q::Union(Node, Type{None}), k::Number)
        r = Node(k)
        if p != None
            r.next = p.next
            if r.next != None
                r.next.prev = r
            end
            p.next = r
            r.prev = p
        end
        if q != None
            r.below = q
            q.above = r
        end
        r
    end

    p = SkipSearch(skip, k)
    q = None
    i = -1
    while true
        i += 1
        if i >= skip.h
            skip.h += 1
            t = skip.s.next
            skip.s = insertAfterAbove(None, skip.s, -Inf)
            insertAfterAbove(skip.s, t, Inf)
        end
        q = insertAfterAbove(p,q,k)
        while p.above == None
            p = p.prev
        end
        p = p.above

        if randbool()
            break
        end
    end
    skip.n += 1
    true
end

function delete(skip::SkipList, k::Number)
    # returns a node p in the bottom of the Skip List such that
    # p.key is the largest key <= k
    function SkipSearch(skip::SkipList, k::Number)
        p = skip.s
        while p.below != None
            p = p.below
            while k >= p.next.key
                p = p.next
            end
        end
        p
    end

    q = SkipSearch(skip, k)
    if q.key != k
        error("$k is not in the skip list")
    end
    while q != None
        p = q.prev
        r = q.next
        p.next = r
        r.prev = p
        q = q.above
    end
    skip.n -= 1
    
    while skip.s.below.next.key == Inf
        skip.s = skip.s.below
    end
    true
end

show(io::IO, node::Node) = print("Node(", node.key, ")")

function show(io::IO, skip::SkipList)
    if skip.n <= 32
        p = skip.s
        while p != None
            t = p
            while t != None
                if t.next != None
                    print(t.key, " ->  ")
                else
                    print(t.key)
                end
                t = t.next
            end
            if p.below != None
                println("\n")
            end
            p = p.below
        end
    else
        print("SkipList with ", skip.n, " elements")
    end
end
